import numpy as np 
from scipy.optimize import least_squares 
from scipy.optimize import minimize
from scipy.linalg import lstsq as sclstsq
import scipy.linalg as lin

def harmonicEuler ( sN, fs, t, f0, k1, kN, ks ): 
    """
    Performs inverse calculation of harmonics contaminating a signal. 
    Args:
        sN = signal containing noise 
        fs = sampling frequency
        t = time samples 
        f0 = base frequency of the sinusoidal noise 
        nK = number of harmonics to calculate 

    """
    KK = np.arange(k1, kN+1, 1/ks )
    nK = len(KK)
    A = np.exp(1j* np.tile(KK,(len(t), 1)) * 2*np.pi* (f0/fs) * np.tile(np.arange(1, len(t)+1, 1),(nK,1)).T)

    v = np.linalg.lstsq(A, sN, rcond=None) 
    alpha = np.real(v[0]) 
    beta  = np.imag(v[0]) 

    amp = np.abs(v[0])     
    phase = np.angle(v[0]) 

    h = np.zeros(len(t))
    for ik, k in enumerate(KK):
        h +=  2*amp[ik] * np.cos( 2.*np.pi*(k) * (f0/fs) * np.arange(1, len(t)+1, 1 )  + phase[ik] )
    
    return sN-h

def harmonicNorm (f0, sN, fs, t, k1, kN, ks): 
    #return np.linalg.norm( harmonicEuler(sN, fs, t, f0, k1, kN, ks)) 
    ii =  sN < (3.* np.std(sN))
    return np.linalg.norm( harmonicEuler(sN, fs, t, f0, k1, kN, ks)[ii] ) 

def minHarmonic(sN, fs, t, f0, k1, kN, ks, Bounds, Nsearch):
    
    kNs = kN
    if Nsearch != False:
        kNs = k1+Nsearch    
    if Bounds == 0:
        # CG, BFGS, Newton-CG, L-BFGS-B, TNC, SLSQP, dogleg, trust-ncg, trust-krylov, trust-exact and trust-constr
        res = minimize(harmonicNorm, np.array((f0)), args=(sN, fs, t, k1, kNs, ks), jac='2-point', method='BFGS') # hess=None, bounds=None )
        print("UNbounded search from ", k1, " to ", kNs, res.x[0]) # for f0 with fN=10 in search", f0)
    
    else:
        bnds = ( (f0-Bounds, f0+Bounds), )
        res = minimize(harmonicNorm, (f0,), args=(sN, fs, t, k1, kNs, ks), jac='2-point', method='L-BFGS-B', bounds=bnds ) # hess=None, bounds=None )
        print("bounded ( +-", Bounds, ") search from ", k1, "to", kNs, res.x[0]) # for f0 with fN=10 in search", f0)

    return harmonicEuler(sN, fs, t, res.x[0], k1, kN, ks), res.x[0]#[0]

def harmonicEuler2 ( sN, fs, t, f0, f0k1, f0kN, f0ks, f1, f1k1, f1kN, f1ks ): 
    """
    Performs inverse calculation of harmonics contaminating a signal. 
    Args:
        sN = signal containing noise 
        fs = sampling frequency
        t = time samples 
        f0 = first base frequency of the sinusoidal noise 
        f0k1 = First harmonic to calula11te for f0 
        f0kN = Last base harmonic to calulate for f0
        f0ks = subharmonics to calculate 
        f1 = second base frequency of the sinusoidal noise 
        f1k1 = First harmonic to calulate for f1
        f1kN = Last base harmonic to calulate for f1
        f1ks = subharmonics to calculate at f1 base frequency
    """
    KK0 = np.arange(f0k1, f0kN+1, 1/f0ks)
    nK0 = len(KK0)
    A0 = np.exp(1j* np.tile(KK0,(len(t), 1)) * 2*np.pi* (f0/fs) * np.tile( np.arange(1, len(t)+1, 1), (nK0,1)).T)

    KK1 = np.arange(f1k1, f1kN+1, 1/f1ks)
    nK1 = len(KK1)
    A1 = np.exp(1j* np.tile(KK1,(len(t), 1)) * 2*np.pi* (f1/fs) * np.tile( np.arange(1, len(t)+1, 1),(nK1,1)).T)

    A = np.concatenate((A0, A1), axis=1)

    v = np.linalg.lstsq(A, sN, rcond=None) # rcond=None) #, rcond=1e-8)
    
    amp0 = np.abs(v[0][0:nK0])     
    phase0 = np.angle(v[0][0:nK0]) 
    amp1 = np.abs(v[0][nK0::])     
    phase1 = np.angle(v[0][nK0::]) 

    h = np.zeros(len(t))
    for ik, k in enumerate(KK0):
        h +=  2*amp0[ik] * np.cos( 2.*np.pi*(k) * (f0/fs) * np.arange(1, len(t)+1, 1 ) + phase0[ik] )
    for ik, k in enumerate(KK1):
        h +=  2*amp1[ik] * np.cos( 2.*np.pi*(k) * (f0/fs) * np.arange(1, len(t)+1, 1 ) + phase1[ik] )

    return sN-h

def harmonic2Norm (f0, sN, fs, t, f0k1, f0kN, f0ks, f1k1, f1kN, f1ks): 
    #return np.linalg.norm(harmonicEuler2(f0[0], f0[1], sN, fs, nK, t))
    ii =  sN < (3.* np.std(sN))
    return np.linalg.norm( harmonicEuler2(sN, fs, t, f0[0], f0k1, f0kN, f0ks, f0[1], f1k1, f1kN, f1ks)[ii] ) 

def minHarmonic2(sN, fs, t, f0, f0k1, f0kN, f0ks, f1, f1k1, f1kN, f1ks, Bounds, Nsearch):

    kNs0 = f0kN
    kNs1 = f1kN
    if Nsearch != False:
        kNs0 = f0k1+Nsearch    
        kNs1 = f1k1+Nsearch    

    if Bounds == 0:

        # CG, BFGS, Newton-CG, L-BFGS-B, TNC, SLSQP, dogleg, trust-ncg, trust-krylov, trust-exact and trust-constr
        print("2 UNbounded ( +-", Bounds,") search length ", kNs0, kNs1 ,"for f0", f0, f1)
        res = minimize(harmonic2Norm, np.array((f0, f1)), args=(sN, fs, t, f0k1, kNs0, f0ks, f1k1, kNs1, f1ks), jac='2-point', method='BFGS') # hess=None, bounds=None )
    else:
        # Bounded
        bnds = ( (f0-Bounds, f0+Bounds),(f1-Bounds, f1+Bounds) )
        print("2 bounded ( +-", Bounds,") search length ", kNs0, kNs1 ,"for f0", f0, f1)
        # L-BFGS-B hess=None, bounds=None )
        res = minimize(harmonic2Norm, ((f0,f1)), args=(sN, fs, t, f0k1, kNs0, f0ks, f1k1, kNs1, f1ks), jac='2-point', method='L-BFGS-B', bounds=bnds ) 

    return harmonicEuler2(sN, fs, t, res.x[0], f0k1, f0kN, f0ks, res.x[1], f1k1, f1kN, f1ks), res.x[0], res.x[1]#[0]

def guessf0( sN, fs ):
    S = np.fft.fft(sN)
    w = np.fft.fftfreq( len(sN), 1/fs )
    imax = np.argmax( np.abs(S) )

    #np.save( "sN.npy", S )
    #np.save( "w.npy", w )
    #exit()
    #plt.plot( w, np.abs(S) )
    #plt.show()
    #print(w)
    #print ( w[imax], w[imax+1] )esta bien in english
    return abs(w[imax])

if __name__ == "__main__":
    
    import matplotlib.pyplot as plt 

    f0 = 60      # Hz
    f1 = 60      # Hz
    delta  = np.random.rand() - .5 
    delta2 = np.random.rand() - .5 
    print("delta", delta)
    print("delta2", delta2)
    fs = 10000   # GMR 
    t = np.arange(0, 1, 1/fs)
    phi =  2.*np.pi*np.random.rand() - np.pi 
    phi2 = 2.*np.pi*np.random.rand() - np.pi
    print("phi", phi, phi2)
    A =  1.0
    A2 = 0.0 
    A3 = 1.0 
    nK = 10
    T2 = .200
    sN  = A *np.sin( ( 1*(delta  +f0))*2*np.pi*t + phi ) + \
          A2*np.sin( ( 1*(delta2 +f1))*2*np.pi*t + phi2 ) + \
              np.random.normal(0,.1,len(t)) + \
              + A3*np.exp( -t/T2  ) 

    sNc = A *np.sin(  (1*(delta +f0))*2*np.pi*t + phi ) + \
          A2*np.sin(  (1*(delta2+f1))*2*np.pi*t + phi2 ) + \
              + A3*np.exp( -t/T2  ) 


    guessf0(sN, fs)

    # single freq
    #h = harmonicEuler( f0, sN, fs, nK, t) 
    h = minHarmonic( f0, sN, fs, nK, t) 
    
    # two freqs 
    #h = minHarmonic2( f0+1e-2, f1-1e-2, sN, fs, nK, t) 
    #h = harmonicEuler2( f0, f1, sN, fs, nK, t) 

    plt.figure()
    plt.plot(t, sN, label="sN")
    #plt.plot(t, sN-h, label="sN-h")
    plt.plot(t, h, label='h')
    plt.title("harmonic")
    plt.legend()

    plt.figure()
    plt.plot(t, sN-sNc, label='true noise')
    plt.plot(t, h, label='harmonic removal')
    plt.plot(t, np.exp(-t/T2), label="nmr")
    plt.legend()
    plt.title("true noise")
    
    plt.show()





import os, sys
import numpy as np
from ruamel import yaml

import pyLemma.LemmaCore as lc 
import pyLemma.Merlin as mrln 
import pyLemma.FDEM1D as em1d 

import numpy as np

import matplotlib.pyplot as plt 
import seaborn as sns
sns.set(style="ticks")

from ruamel import yaml

#import cmocean 
#from SEGPlot import *
#from matplotlib.ticker import FormatStrFormatter
#import matplotlib.ticker as plticker


# Converts Lemma/Merlin/Akvo serialized Eigen arrays into numpy ones for use by Python 
class VectorXr(yaml.YAMLObject):
    """
    Converts Lemma/Merlin/Akvo serialized Eigen arrays into numpy ones for use by Python 
    """
    yaml_tag = u'VectorXr'
    def __init__(self, array):
        self.size = np.shape(array)[0]
        self.data = array.tolist()
    def __repr__(self):
        # Converts to numpy array on import 
        return "np.array(%r)" % (self.data)

class AkvoData(yaml.YAMLObject):
    """
    Reads an Akvo serialized dataset into a standard python dictionary 
    """
    yaml_tag = u'AkvoData'
    def __init__(self, array):
        pass
        #self.size = np.shape(array)[0]
        #self.Imp = array.tolist()
    def __repr__(self):
        # Converts to a dictionary with Eigen vectors represented as Numpy arrays 
        return self

def loadAkvoData(fnamein):
    """ Loads data from an Akvo YAML file. The 0.02 is hard coded as the pulse length. This needs to be 
        corrected in future kernel calculations. The current was reported but not the pulse length. 
    """
    fname = (os.path.splitext(fnamein)[0])
    with open(fnamein, 'r') as stream:
        try:
            AKVO = (yaml.load(stream, Loader=yaml.Loader))
        except yaml.YAMLError as exc:
            print(exc)
    return AKVO 


def main():

    if len(sys.argv) < 3:
        print ("usage  python calcAkvoKernel.py   AkvoDataset.yaml  Coil1.yaml kparams.yaml  SaveString.yaml " )
        exit()

    AKVO = loadAkvoData(sys.argv[1])

    B_inc = AKVO.META["B_0"]["inc"]  
    B_dec = AKVO.META["B_0"]["dec"]  
    B0    = AKVO.META["B_0"]["intensity"]  

    fT = AKVO.transFreq
    #gamma = 2.67518e8
    #B0 = (fL*2.*np.pi) /gamma * 1e9
 
    Coil1 = em1d.PolygonalWireAntenna.DeSerialize( sys.argv[2] )
    Coil1.SetNumberOfFrequencies(1)
    Coil1.SetFrequency(0, fT) 
    Coil1.SetCurrent(1.)

    # read in kernel params
    kparams = loadAkvoData( sys.argv[3] )

    ## TODO 
    # pass this in...
    lmod = em1d.LayeredEarthEM() 

    nlay = len(kparams["sigs"])
    sigs = np.array(kparams["sigs"])
    tops = np.array(kparams["tops"])
    bots = np.array(kparams["bots"])
    
    if ( (len(tops)-1) != len(bots)):
        print("Layer mismatch")
        exit()

    thicks = bots - tops[0:-1]
    
    lmod.SetNumberOfLayers(nlay + 1)
    lmod.SetLayerThickness(thicks)
    lmod.SetLayerConductivity( np.concatenate( ( [0.0], sigs ) ))

    #lmod.SetNumberOfLayers(4)
    #lmod.SetLayerThickness([15.49, 28.18])
    #lmod.SetLayerConductivity([0.0, 1./16.91, 1./24.06, 1./33.23])


    lmod.SetMagneticFieldIncDecMag( B_inc, B_dec, B0, lc.NANOTESLA )

    Kern = mrln.KernelV0()
    Kern.PushCoil( "Coil 1", Coil1 )
    Kern.SetLayeredEarthEM( lmod );
    Kern.SetIntegrationSize( (kparams["size_n"], kparams["size_e"], kparams["size_d"]) )
    Kern.SetIntegrationOrigin( (kparams["origin_n"], kparams["origin_e"], kparams["origin_d"]) )
    Kern.SetTolerance( 1e-9*kparams["branchTol"] )
    Kern.SetMinLevel( kparams["minLevel"] )
    Kern.SetMaxLevel( kparams["maxLevel"] )
    Kern.SetHankelTransformType( lc.FHTKEY201 )
    Kern.AlignWithAkvoDataset( sys.argv[1] )

    if str(kparams["Lspacing"]).strip() == "Geometric":
        thick = np.geomspace(kparams["thick1"], kparams["thickN"], num=kparams["nLay"])
    elif str(kparams["Lspacing"]) == "Log":
        thick = np.logspace(kparams["thick1"], kparams["thickN"], num=kparams["nLay"])
    elif str(kparams["Lspacing"]) == "Linear":
        thick = np.linspace(kparams["thick1"], kparams["thickN"], num=kparams["nLay"])
    else:
        print("DOOOM!, in calcAkvoKernel layer spacing was not <Geometric>, <Log>, or <Linear>")
        print( str(kparams["Lspacing"]) )
        exit()

    iface = np.cumsum(thick)
    Kern.SetDepthLayerInterfaces(iface)
    #Kern.SetDepthLayerInterfaces(np.geomspace(1, 110, num=40))
    #Kern.SetDepthLayerInterfaces(np.linspace(1, 110, num=50))
    #Kern.SetDepthLayerInterfaces(np.geomspace(1, 110, num=40))
 
    # autAkvoDataNode = YAML::LoadFile(argv[4]);
    # Kern->AlignWithAkvoDataset( AkvoDataNode );

    Kern.CalculateK0( ["Coil 1"], ["Coil 1"], False )

    #yml = open( 'test' + str(Kern.GetTolerance()) + '.yaml', 'w')
    yml = open( sys.argv[4], 'w' )
    print(Kern, file=yml)

    # 
    K0 = Kern.GetKernel()
    plt.matshow(np.abs(K0))
    plt.show()

if __name__ == "__main__":
    main()



import numpy as np
from numpy.linalg import lstsq
from numpy.linalg import norm
from numpy import fft 

import pylab

from scipy.signal import correlate

def autocorr(x):
    #result = np.correlate(x, x, mode='full')
    result = correlate(x, x, mode='full')
    return result[result.size/2:]

class AdaptiveFilter:

    def __init__(self, mu):
        self.mu = mu

    def adapt_filt_Ref(self, x, R, M, mu, PCA, lambda2=0.95, H0=0):
        """ Taken from .m file
        This function is written to allow the user to filter a input signal   
        with an adaptive filter that utilizes 2 reference signals instead of  
        the standard method which allows for only 1 reference signal.         
        Author: Rob Clemens              Date: 3/16/06                       
        Modified and ported to Python, now takes arbitray number of reference points 
        Original public domain source 
        https://www.mathworks.com/matlabcentral/fileexchange/10447-noise-canceling-adaptive-filter 
            x = data array 
            R = reference array 
            M = number of taps 
            mu = forgetting factor 
            PCA = Perform PCA 
        """
        #from akvo.tressel import pca 
        import akvo.tressel.pca as pca 

        if np.shape(x) != np.shape(R[0]): # or np.shape(x) != np.shape(rx1):
            print ("Error, non aligned")
            exit(1)
        
        if PCA == "Yes":
            #print("Performing PCA calculation in noise cancellation")
            # PCA decomposition on ref channels so signals are less related
            R, K, means = pca.pca( R )
           
            # test for in loop reference  
            #print("Cull nearly zero terms?", np.shape(x), np.shape(R))     
            #R = R[0:3,:] 
            #R = R[2:4,:] 
            #print("    removed zero terms?", np.shape(x), np.shape(R))     
            #H0 = H0[0:3*np.shape(x)[0]]
            #H0 = H0[0:2*np.shape(x)[0]]

        if all(H0) == 0:
            # corrects for dimensionality issues if a simple 0 is passed
            H = np.zeros( (len(R)*M))
        else:
            H = H0

        Rn = np.ones(len(R)*M) / mu 
        
        r_ = np.zeros( (len(R), M) ) 
        e = np.zeros(len(x))           # error, in our case the desired output
        ilambda = lambda2**-1

        for ix in range(0, len(x)):
            # Only look forwards, to avoid distorting the lates times 
            # (run backwards, if opposite and you don't care about distorting very late time.)
            for ir in range(len(R)):  # number of reference channels 
                if ix < M:
                    r_[ir,0:ix] = R[ir][0:ix]
                    r_[ir,ix:M] = 0 
                else:
                    r_[ir,:] = R[ir][ix-M:ix]

            # reshape            
            r_n = np.reshape(r_, -1) # concatenate the ref channels in to a 1D array 

            K      = (Rn* r_n) / (lambda2 + np.dot(r_n*Rn, r_n))       # Create/update K
            e[ix]  = x[ix] - np.dot(r_n.T, H)                          # e is the filtered signal, input - r(n) * Filter Coefs
            H     += K*e[ix];                                          # Update Filter Coefficients
            Rn     = ilambda*Rn - ilambda*np.dot(np.dot(K, r_n.T), Rn) # Update R(n)
        return e, H

    def transferFunctionFFT(self, D, R, reg=1e-2):
         from akvo.tressel import pca
         """
             Computes the transfer function (H) between a Data channel and 
             a number of Reference channels. The Matrices D and R are 
             expected to be in the frequency domain on input.
             | R1'R1 R1'R2 R1'R3|   |h1|   |R1'D|
             | R2'R1 R2'R2 R2'R3| * |h2| = |R2'D|
             | R3'R1 R3'R2 R3'R3|   |h3|   |R3'D|
 
             Returns the corrected array 
         """
 
         # PCA decomposition on ref channels so signals are less related
         #transMatrix, K, means = pca.pca( np.array([rx0, rx1]))   
         #RR = np.zeros(( np.shape(R[0])[0]*np.shape(R[0])[1], len(R)))
 #         RR = np.zeros(( len(R), np.shape(R[0])[0]*np.shape(R[0])[1] ))
 #         for ir in range(len(R)):
 #             RR[ir,:] = np.reshape(R[ir], -1)
 #         transMatrix, K, means = pca.pca(RR)    
 #         #R rx0 = transMatrix[0,:]
 #         # rx1 = transMatrix[1,:]
 #         for ir in range(len(R)):
 #             R[ir] = transMatrix[ir,0]
 
         import scipy.linalg 
         import akvo.tressel.pca as pca 
         # Compute as many transfer functions as len(R)
         # A*H = B
         nref = len(R)
         H = np.zeros( (np.shape(D)[1], len(R)), dtype=complex )
         for iw in range(np.shape(D)[1]):
             A = np.zeros( (nref, nref), dtype=complex )
             B = np.zeros( (nref) , dtype=complex)
             for ii in range(nref):
                 for jj in range(nref):
                     # build A
                     A[ii,jj] = np.dot(R[ii][:,iw], R[jj][:,iw])                 
 
                 # build B
                 B[ii] = np.dot( R[ii][:,iw], D[:,iw] )
 
             # compute H(iw)
             #linalg.solve(a,b) if a is square
             #print "A", A
             #print "B", B
             # TODO, regularise this solve step? So as to not fit the spurious noise
             #print np.shape(B), np.shape(A) 
             #H[iw, :] = scipy.linalg.solve(A,B)
             H[iw, :] = scipy.linalg.lstsq(A,B,cond=reg)[0]
             #print "lstqt", np.shape(scipy.linalg.lstsq(A,B))
             #print "solve", scipy.linalg.solve(A,B)
             #H[iw,:]  = scipy.linalg.lstsq(A,B) # otherwise 
                 #H = np.zeros( (np.shape(D)[1], )   )
         #print H #A, B
         Error = np.zeros(np.shape(D), dtype=complex)
         for ir in range(nref):
             for q in range( np.shape(D)[0] ):
                 #print "dimcheck", np.shape(H[:,ir]), np.shape(R[ir][q,:] )
                 Error[q,:] += H[:,ir]*R[ir][q,:]
         return D - Error




def plotLogo(ax):
    #import matplotlib.pyplot as ax 
    import numpy as np 

    x  = np.arange(-np.pi/2, 7.0*np.pi, .01)
    y  = np.sin(x)
    y2 = np.cos(x)

    y3 = np.sin(.5*x)
    y4 = np.cos(.5*x)

    ax.plot(x,  y, linewidth=1, color='lightgrey')
    ax.plot(x, -y, linewidth=1, color='lightgrey')
    ax.plot(x, y2, linewidth=1, color='lightgrey')
    ax.plot(x,-y2, linewidth=1, color='lightgrey')

    ax.plot(x,  y3, linewidth=1, color='lightgrey')
    ax.plot(x, -y3, linewidth=1, color='lightgrey')
    ax.plot(x,  y4, linewidth=1, color='lightgrey')
    ax.plot(x, -y4, linewidth=1, color='lightgrey')

    a = np.arange(10,615)
    ax.plot(x[a],y[a], linewidth=3, color='black')
    ax.fill_between(x[a],y[a], y2=-1, where=y[a]>=-1, interpolate=True, linewidth=0, alpha=.95, color='maroon')

    k  = np.arange(615, 785)
    k2 = np.arange(785, 1105)
    ax.plot(x[k],    y[k ], linewidth=3, color='black')
    ax.plot(x[k],   -y[k ], linewidth=3, color='black')
    ax.plot(x[k2],  y3[k2], linewidth=3, color='black')
    ax.plot(x[k2], -y3[k2], linewidth=3, color='black')

    #v = np.arange(1265,1865) # y 
    v = np.arange(1105, 1720)
    ax.plot(x[v], -y2[v], linewidth=3, color='black')

    o = np.arange(1728,2357)
    ax.plot(x[o],  y4[o], linewidth=3, color='black')
    ax.plot(x[o], -y4[o], linewidth=3, color='black')
    ax.fill_between(x[o],  y4[o], y2=-y4[o],  where=y4[o]<=1, interpolate=True, linewidth=0, alpha=.95, color='maroon')

if __name__ == "__main__":
    import matplotlib.pyplot as plt 

    fig = plt.figure( figsize=(6,3) )
    #ax = fig.add_axes([.1,.1,.8,.8])
    ax = fig.add_subplot(211)

    fig.patch.set_facecolor( None )
    fig.patch.set_alpha( .0 )
    ax.axis('off')
    plotLogo(ax)
    ax.xaxis.set_major_locator(plt.NullLocator()) 
    ax.yaxis.set_major_locator(plt.NullLocator()) 

    subplot2 = fig.add_subplot(212)
    subplot2.text(0.5, 1.,'surface NMR workbench',
            horizontalalignment='center',
            verticalalignment='center',
            size=22,
            transform = subplot2.transAxes)
    subplot2.xaxis.set_major_locator(plt.NullLocator()) 
    subplot2.yaxis.set_major_locator(plt.NullLocator()) 
    subplot2.axis('off')
    plt.savefig("logo.pdf")
    plt.show()

#ax.fill_between(x[o], -y4[o], y2=0, where=-y4[o]<=1, interpolate=True, linewidth=0, alpha=.5, color='black')
#ax.plot(x[o], y2[o], linewidth=3, color='black')
#ax.plot(x[o],-y2[o], linewidth=3, color='black')
#ax.fill_between(x[a], y[a], y2=-1, where=y[a]>=-1, interpolate=True, linewidth=0, alpha=.5, color='black')

#ax.show()



from __future__ import division 

import matplotlib as mpl
mpl.use('pdf')

#from rasterize import rasterize_and_save

import matplotlib.patches as mpatches
from pwctime import pwcTime
from logbarrier import * 
from perlin import perlin
from scipy import stats
import cmocean
import sys 
import numpy as np 
import seaborn as sns

def bootstrapWindows(N, nboot, isum, adapt=False):
    """ Bootstraps noise as a function of gate width
        N = input noise signal 
        nboot = number of boostrap windows to perform 
        isum = length of windows (L_i)
        adapt = reduce nboot as window size increases
    """
    nc = np.shape(N)[0]
    Means = {}

    if adapt:
        Means = -9999*np.ones((len(isum), nboot//isum[0])) # dummy value
        for ii, nwin in enumerate(isum):  
            for iboot in range(nboot//isum[ii]):
                cs = np.random.randint(0,nc-nwin)
                Means[ii,iboot] = np.mean( N[cs:cs+nwin] )
        Means = np.ma.masked_less(Means, -9995)

    else:
        Means = np.zeros((len(isum), nboot))
        for ii, nwin in enumerate(isum):  
            for iboot in range(nboot):
                cs = np.random.randint(0,nc-nwin)
                Means[ii,iboot] = np.mean( N[cs:cs+nwin] )

    return Means, np.array(isum)

def gateIntegrate(T2D, T2T, gpd, sigma, stackEfficiency=2.):
    """ Gate integrate the signal to gpd, gates per decade
        T2D = the time series to gate integrate, complex 
        T2T = the abscissa values 
        gpd = gates per decade 
        sigma = estimate of standard deviation for theoretical gate noise 
        stackEfficiency = exponential in theoretical gate noise, 2 represents ideal stacking
    """
    
    # use artificial time gates so that early times are fully captured
    T2T0 = T2T[0]
    T2TD = T2T[0] - (T2T[1]-T2T[0])
    T2T -= T2TD
    
    #####################################
    # calculate total number of decades #
    # windows edges are approximate until binning but will be adjusted to reflect data timing, this 
    # primarily impacts bins with a few samples  
    nd = np.log10(T2T[-1]/T2T[0])               
    tdd = np.logspace( np.log10(T2T[0]), np.log10(T2T[-1]), (int)(gpd*nd)+1, base=10, endpoint=True) 
    tdl = tdd[0:-1]                 # approximate window left edges
    tdr = tdd[1::]                  # approximate window right edges
    td = (tdl+tdr) / 2.             # approximate window centres


    Vars = np.zeros( len(td) ) 
    htd = np.zeros( len(td), dtype=complex )
    isum = np.zeros( len(td), dtype=int )  

    ii = 0
    for itd in range(len(T2T)):
        if ( T2T[itd] > tdr[ii] ):
            ii += 1
            # correct window edges to centre about data 
            tdr[ii-1] = (T2T[itd-1]+T2T[itd])*.5 
            tdl[ii  ] = (T2T[itd-1]+T2T[itd])*.5
        isum[ii] += 1
        htd[ii] += T2D[ itd ]
        Vars[ii] += sigma**2
        
    td = (tdl+tdr) / 2.             # actual window centres
    sigma2 = np.sqrt( Vars * ((1/(isum))**stackEfficiency) ) 

    # Reset abscissa where isum == 1 
    # when there is no windowing going on 
    td[isum==1] = T2T[0:len(td)][isum==1]

    tdd = np.append(tdl, tdr[-1])

    htd /= isum # average
    T2T += T2TD # not used  
    return td+T2TD, htd, tdd+T2TD, sigma2, isum  # centre abscissa, data, window edges, error 

PhiD = []
def invert(Time, t, v, sig, lambdastar):
    """ helper function that simply calls logBarrier, here to allow for drop in repacement  
    """
    #model = logBarrier(Time.Genv, 1e-2*v, Time.T2Bins, MAXITER=5000, sigma=1e-2*sig, alpha=1e6, smooth="Both") 
    model = logBarrier(Time.Genv, 1e-2*v, Time.T2Bins, lambdastar, MAXITER=750, sigma=1e-2*sig, alpha=1e6, smooth="Smallest") 
    PhiD.append(model[2])
    return model

def gateTest(vc, vgc, pperlin, boot, lamdastar):
    """ Performs gate integration and adds random noise 
        vc = clean data (dense)
        vgc = clean data at gates 
        boot = if "boot" then bootstrap the gate noise 
        lambdastar = l-curve or discrepency principle
        pperlin = percent perlin noise, noise floor is maintained at 2.00 PU 
    """

    t = np.arange(2e-4, .3601, 2e-4)
    zeta = np.pi / 3.
    
    v = np.copy(vc) # important!  

    # Scaling factors to keep noise floor constant with increasing levels of 
    # Perlin noise. These were determined using populations of 5,000 and hold to 
    # two significant digits (i.e, 2.00 PU) 
    PF = {0.0:0,\
          2.5:.450,\
          5.0:.6125,\
          7.5:.765,\
          10.0:.87375,\
          12.5:.9725,\
          15.0:1.05,\
          17.5:1.1275,\
          20.0:1.20,\
          22.5:1.265,\
          25.0:1.325}
 
    # random noise
    np.random.seed() # necessary for thread pool, otherwise all threads can get same numbers 
    sigma = 2.*(1.-1e-2*pperlin)
    eps = np.random.normal(0, sigma, len(t)) + \
       1j*np.random.normal(0, sigma, len(t)) 
    eps += PF[pperlin] * perlin(len(t), L=.3601, sigma_f=.005, sigma_r=0.72) # 1 PU std
    v += eps

    # Noise based on residual  
    sigmahat = np.std( v.imag )
    
    gt, gd, we, err, isum = gateIntegrate(v.real, t, 20, sigmahat)
    ge = np.copy(err)
    if boot=="boot":
        Means, isum2 = bootstrapWindows(v.imag, 20000, isum[isum!=1], adapt=True)
        # STD 
        #err[isum!=1] = np.ma.std(Means, axis=1, ddof=1)[isum!=1]
        # MAD, only for windows > 1 
        c = stats.norm.ppf(3./4.)
        err[isum!=1] = np.ma.median(np.ma.abs(Means), axis=1) / c 
    if boot=="uniform":
        err = sigmahat
 
    return gt, gd, gd-vgc, err, v.real, t, isum

if __name__ == "__main__":

    import multiprocessing 
    import itertools 

    from GJIPlot import *
    import numpy as np 
    import matplotlib.pyplot as plt 
    from matplotlib.ticker import FormatStrFormatter
    from matplotlib import ticker
    from collections import OrderedDict

    if len(sys.argv)<4:
        print ( "Python script for generating plots used in GJI publication")
        print ( "useage:")
        print ( "python gateIntegrate4.py   NoiseType   Sigma_i  Lambda*   " )
        exit()

    if sys.argv[1] not in ['0.0','2.5','5.0','7.5','10.0','12.5','15.0','17.5','20.0','22.5','25.0']: 
        print ( "PercentPerlin: [0.0,2.5,5.0...25.0] ", "got", sys.argv[1])
        exit(1)

    if sys.argv[2] != "gauss" and sys.argv[2] != "boot" and sys.argv[2] != "uniform":
        print ( "Sigma_i: gauss | boot | uniform")
        exit(1)
    
    if sys.argv[3] != "lcurve" and sys.argv[3] != "discrepency": 
        print ( "Lambda*: lcurve | discrepency ")
        exit(1)

    #offwhite = (.98,.98,.98)
    offwhite = (1.,1.,1.) 
    mDarkBrown = '#eb811b' # alert colour 
    mDarkTeal = '#23373b'
    mLightBrown= "#EB811B"
    mLightGreen = "#14B03D"

    # Time series plot
    fig = plt.figure(figsize=(pc2in(18),pc2in(18)), facecolor=offwhite)
    ax2  = fig.add_axes([.195, .175, .750, .75], facecolor=offwhite)  # time
    
    # Main plot 
    fig2 = plt.figure(figsize=(pc2in(20),pc2in(2*.5*20)))
    ax1  = fig2.add_axes([.175, .410*1.5, .6,   .225*1.5])
    ax1c = fig2.add_axes([.800, .410*1.5, .025, .225*1.5])
    ax3  = fig2.add_axes([.175, .100*1.5, .495, .225*1.5], facecolor='None')
    ax3r  = fig2.add_axes([.175, .100*1.5, .495, .225*1.5], facecolor='None', rasterized=True, sharex=ax3, sharey=ax3)
    ax3b = fig2.add_axes([.825, .100*1.5, .1,   .225*1.5])

    SIG = []
    ER = []
    GD = []
    GT = []
    V = []
    MOD = []
    CONV = []
    PHID = []
    PHIM = []
    LSTAR = []
    ns = 10000 #10000  #10000 # number of realizations for PDF 
    ni = 5000  #5000  #1000  # number of inversions to plot 
    t = np.arange(2e-4, .3601, 2e-4) # CMR sampling
 
    #CMAP = cmocean.cm.solar
    CMAP = cmocean.cm.gray_r
    #CMAP = cmocean.cm.haline
    #CMAP = cmocean.cm.tempo

    ##############################################
    # set up model 
    lowT2 = .001
    hiT2 = 1.0 
    nT2 = 30
    spacing = "Log_10"
    Time = pwcTime()
    Time.setT2(lowT2, hiT2, nT2, spacing) 
    Time.setSampling( np.arange(2e-4, .3601, 2e-4) )
    Time.generateGenv()
    tmod = np.zeros(nT2)
    tmod [8] = .15  # distribution centres...to be smoothed
    tmod [20] = .1
    for i in range(2):
        tmod = np.convolve(tmod, np.array([.0625,.125,.1875,.25,.1875,.125,.0625]), 'same') 

    vc = 100. * np.dot(Time.Genv, tmod) + 0j # in PU
    gt, gd, we, err, isum = gateIntegrate(vc, t, 20, 3)

    ##############################################
    # Set up inversion 
    Time = pwcTime()
    Time.setT2(lowT2, hiT2, nT2, spacing) 
    Time.setSampling( gt ) 
    Time.generateGenv()
    vgc = 100.*np.dot(Time.Genv, tmod) + 0j # in PU

    # make the Pool of workers
    print("pool gate integrate")
    with multiprocessing.Pool() as pool: 
        results = pool.starmap(gateTest, zip(np.tile(vc, (ns, 1)), np.tile(vgc, (ns,1)), itertools.repeat(eval(sys.argv[1])), \
                                                                                         itertools.repeat(sys.argv[2]), \
                                                                                         itertools.repeat(sys.argv[3])))
    print("done pool gate integrate")
   
    # parse out results 
    for i in range(ns):
        gt,gd,ge,err,v,vt,isum = results[i] 
        V.append(v.real)
        GT.append(gt.real)
        GD.append(gd.real)
        ER.append( ge.real / err.real )
        SIG.append( err.real )

    print("pool inversions")
    with multiprocessing.Pool() as pool: 
        invresults = pool.starmap(invert, zip(itertools.repeat(Time), GT[0:ni], GD[0:ni], SIG[0:ni], itertools.repeat(sys.argv[3]) )) 
    #print("done pool inversions",results[:][0])

    # Parse results 
    for i in range(ns):
        #print("Sym %", round(100.*i/(float)(1)/(float)(ns)))
        # invert
        if i < ni:
            #mod, conv, phid = invert(Time, gt, gd.real, err)
            if sys.argv[3] == "discrepency":
                mod, conv, phid_final = invresults[i] 
            else:
                mod, conv, phid_final, phim, phid, lstar = invresults[i] 
            MOD.append(mod)
            CONV.append(conv)
            PHID.append(phid)
            PHIM.append(phim)
            LSTAR.append(lstar)

    PHIM = np.array(PHIM)
    PHID = np.array(PHID)
    ER = np.array(ER)
    MOD = np.array(MOD)
    GD = np.array(GD)

    ####################
    # Time series plot #       
    ax2.plot( 1e3*vt, V[0], color=mDarkTeal, label="$V_N$", linewidth=1, zorder=-32) #, rasterized=True) 
    ax2.errorbar( 1e3*gt, GD[0], yerr=SIG[0], fmt='.', markersize=6, color=mLightBrown, label="$V_G$")
    ax2.set_ylim([-10,30])
    leg1 = ax2.legend( labelspacing=0.2, scatterpoints=1, numpoints=1, frameon=True )   
    fixLeg(leg1)
    ax2.set_xscale("log", nonposx='clip')  
    ax2.set_ylabel(r"$V_N$ (PU)")
    ax2.get_xaxis().set_major_formatter(FormatStrFormatter('%1.0f'))
    ax2.set_xlabel("time (ms)")
    deSpine(ax2)
    fig.savefig( sys.argv[1] + "-" + sys.argv[2] + "-" + sys.argv[3] + "-ts.pdf", dpi=400, facecolor=offwhite,edgecolor=offwhite)

    # histogram of error statistic
    bins = np.linspace( -3, 3, 40, endpoint=True )
    HIST = []
    for i in range(0,np.shape(ER)[1]):
        hist, edges = np.histogram(ER[:,i], bins=bins, density=False)        
        HIST.append(hist)
    HIST =  np.array(HIST)/(float)(ns) # normalize 
        
    im = ax1.pcolor(1e3*we, edges, HIST.T, cmap=CMAP, vmin=0, vmax=.1, rasterized=True)
    im.set_edgecolor('face')
    cb = plt.colorbar(im, ax1c, label=r"probability density", format=FormatStrFormatter('%1.2f'))
    cb.solids.set_rasterized(True) 
    tick_locator = ticker.MaxNLocator(nbins=4)
    cb.locator = tick_locator
    cb.update_ticks()

    ax1.set_xscale("log", nonposx='clip')
    ax1.get_xaxis().set_major_formatter(FormatStrFormatter('%1.0f'))
    ax1.set_xlabel("time (ms)")
    ax1.set_ylabel(r"gate error $\left( \left( {V_G - V_T} \right) / {\tilde{\sigma_i}} \right)$")

    LMT2 = [] 
    THETA = []
    MODERR = []
       
    # plot a random sample of ns instead?  
    for i in range(ni):
        # plot log mean and amplitude
        model = MOD[i] 
        theta = np.sum( model )
        LogMeanT2 = np.exp(np.sum( model * np.log( Time.T2Bins ) ) / theta ) 
        LMT2.append(LogMeanT2)
        THETA.append( np.sum(model)  )
        MODERR.append( np.linalg.norm(model-tmod) )
                
    CONV = np.array(CONV)
    THETA = np.array(THETA)
    MOD = np.array(MOD)
    MODERR = np.array(MODERR)
        
    #############################
    # plot all models, 1 colour #
    ires = ax3r.plot( 1e3*np.tile(Time.T2Bins, (np.sum(np.array(CONV)) ,1)).T , 1e2*MOD.T, color=mDarkTeal, alpha=.01, lw=.5, label="$\mathbf{f}_I$", zorder=0, rasterized=True) 
    lns2, = ax3r.plot(1e3*Time.T2Bins, 1e2*tmod, color=mLightBrown, linewidth=2, label="$\mathbf{f}_T$")

    handles, labels = ax3r.get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    leg3 = ax3r.legend(by_label.values(), by_label.keys(), labelspacing=0.2, scatterpoints=1, numpoints=1, frameon=True , loc="upper right")            
    for line in leg3.get_lines():
        line.set_linewidth(1)
    for lh in leg3.legendHandles: 
        lh.set_alpha(1)
    fixLeg(leg3)

    ###########################
    # Error histogram on side #
    ax3b.hist( 1e2*MODERR, bins='auto', orientation="horizontal", color=mDarkTeal, stacked=True, density=True, range=(0,20)) 
    ax3b.axhline(1e2*np.mean(MODERR), linewidth=1.25, color=mLightBrown) #, color=CMAP(0.7), zorder=1)
    deSpine(ax3b)
    ax3b.set_xscale("log", nonposx='clip')  
    ax3b.set_ylabel(r"$\Vert \mathbf{f}_I -\mathbf{f}_T \Vert$") # %(m$^3$/m$^3$)") #, color="C0")   
    ax3b.set_xlabel("log probability\ndensity") #, color="C0")   
    ax3.set_xlim( (1e3*Time.T2Bins[0], 1e3*Time.T2Bins[-1]) ) 
    ax3.set_ylim( (0,5) )  
    ax3.set_xlabel("$T_2$ (ms)")   
    ax3.set_ylabel("partial water content (PU)") #, color="C0")   
    ax3.set_xscale("log", nonposx='clip')  
    ax3.get_xaxis().set_major_formatter(FormatStrFormatter('%1.0f'))
    plt.setp(ax3r.get_xticklabels(), visible=False)
    plt.setp(ax3r.get_yticklabels(), visible=False)
    deSpine(ax3)
    deSpine(ax3r)
    
    np.save("pperlin" + str(round(1e1*eval(sys.argv[1]))) + "-" + sys.argv[2] + "-" + sys.argv[3] + "-err", MODERR) 

    plt.savefig("pperlin" + str(round(1e1*eval(sys.argv[1]))) + "-" + sys.argv[2] + "-" + sys.argv[3] + ".pdf", dpi=600, facecolor=offwhite, edgecolor=offwhite)
    
    plt.show()




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/tirons/src/akvo/akvo/gui/addFigure8Loop.ui'
#
# Created by: PyQt5 UI code generator 5.13.1
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_figure8LoopAdd(object):
    def setupUi(self, figure8LoopAdd):
        figure8LoopAdd.setObjectName("figure8LoopAdd")
        figure8LoopAdd.resize(400, 456)
        self.gridLayout = QtWidgets.QGridLayout(figure8LoopAdd)
        self.gridLayout.setObjectName("gridLayout")
        self.doubleSpinBox_3 = QtWidgets.QDoubleSpinBox(figure8LoopAdd)
        self.doubleSpinBox_3.setObjectName("doubleSpinBox_3")
        self.gridLayout.addWidget(self.doubleSpinBox_3, 12, 2, 1, 1)
        self.label_8 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 7, 0, 1, 1)
        self.segments = QtWidgets.QSpinBox(figure8LoopAdd)
        self.segments.setMinimum(5)
        self.segments.setMaximum(200)
        self.segments.setProperty("value", 15)
        self.segments.setObjectName("segments")
        self.gridLayout.addWidget(self.segments, 14, 2, 1, 1)
        self.loopHeight = QtWidgets.QDoubleSpinBox(figure8LoopAdd)
        self.loopHeight.setDecimals(3)
        self.loopHeight.setMinimum(-99.0)
        self.loopHeight.setProperty("value", -0.001)
        self.loopHeight.setObjectName("loopHeight")
        self.gridLayout.addWidget(self.loopHeight, 10, 2, 1, 1)
        self.label_6 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 13, 0, 1, 1)
        self.centreNorth2 = QtWidgets.QDoubleSpinBox(figure8LoopAdd)
        self.centreNorth2.setObjectName("centreNorth2")
        self.gridLayout.addWidget(self.centreNorth2, 6, 2, 1, 1)
        self.line_4 = QtWidgets.QFrame(figure8LoopAdd)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout.addWidget(self.line_4, 8, 1, 1, 1)
        self.line = QtWidgets.QFrame(figure8LoopAdd)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 5, 0, 1, 1)
        self.centreEast1 = QtWidgets.QDoubleSpinBox(figure8LoopAdd)
        self.centreEast1.setMinimum(-99999999.0)
        self.centreEast1.setMaximum(99999999.99)
        self.centreEast1.setObjectName("centreEast1")
        self.gridLayout.addWidget(self.centreEast1, 1, 2, 1, 1)
        self.label_9 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_9.setObjectName("label_9")
        self.gridLayout.addWidget(self.label_9, 12, 0, 1, 1)
        self.label_2 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)
        self.line_3 = QtWidgets.QFrame(figure8LoopAdd)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout.addWidget(self.line_3, 8, 2, 1, 1)
        self.label = QtWidgets.QLabel(figure8LoopAdd)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.label_7 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 6, 0, 1, 1)
        self.label_5 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 14, 0, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(figure8LoopAdd)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 17, 2, 1, 1)
        self.line_2 = QtWidgets.QFrame(figure8LoopAdd)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout.addWidget(self.line_2, 5, 2, 1, 1)
        self.centreEast2 = QtWidgets.QDoubleSpinBox(figure8LoopAdd)
        self.centreEast2.setObjectName("centreEast2")
        self.gridLayout.addWidget(self.centreEast2, 7, 2, 1, 1)
        self.label_3 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 10, 0, 1, 1)
        self.centreNorth1 = QtWidgets.QDoubleSpinBox(figure8LoopAdd)
        self.centreNorth1.setMinimum(-999999999.0)
        self.centreNorth1.setMaximum(9999999999.0)
        self.centreNorth1.setObjectName("centreNorth1")
        self.gridLayout.addWidget(self.centreNorth1, 0, 2, 1, 1)
        self.loopTurns = QtWidgets.QSpinBox(figure8LoopAdd)
        self.loopTurns.setMinimum(1)
        self.loopTurns.setMaximum(100000)
        self.loopTurns.setObjectName("loopTurns")
        self.gridLayout.addWidget(self.loopTurns, 13, 2, 1, 1)
        self.line_5 = QtWidgets.QFrame(figure8LoopAdd)
        self.line_5.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout.addWidget(self.line_5, 8, 0, 1, 1)
        self.label_4 = QtWidgets.QLabel(figure8LoopAdd)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 9, 0, 1, 1)
        self.loopRadius = QtWidgets.QDoubleSpinBox(figure8LoopAdd)
        self.loopRadius.setMinimum(0.1)
        self.loopRadius.setMaximum(600.0)
        self.loopRadius.setProperty("value", 50.0)
        self.loopRadius.setObjectName("loopRadius")
        self.gridLayout.addWidget(self.loopRadius, 9, 2, 1, 1)

        self.retranslateUi(figure8LoopAdd)
        self.buttonBox.accepted.connect(figure8LoopAdd.accept)
        self.buttonBox.rejected.connect(figure8LoopAdd.reject)
        QtCore.QMetaObject.connectSlotsByName(figure8LoopAdd)

    def retranslateUi(self, figure8LoopAdd):
        _translate = QtCore.QCoreApplication.translate
        figure8LoopAdd.setWindowTitle(_translate("figure8LoopAdd", "Dialog"))
        self.label_8.setText(_translate("figure8LoopAdd", "centre easting 2 (m)"))
        self.segments.setToolTip(_translate("figure8LoopAdd", "<html><head/><body><p>Currently Akvo/Merlin calculates circular loops using segments of wire, forming a polygon. Analytic circular loops may be added in the future. </p></body></html>"))
        self.loopHeight.setToolTip(_translate("figure8LoopAdd", "<html><head/><body><p>Akvo uses a positive down convention, a slight negative value to the loop height improves numerical stability using digital filtering hankel transforms. </p></body></html>"))
        self.label_6.setText(_translate("figure8LoopAdd", "turns"))
        self.label_9.setText(_translate("figure8LoopAdd", "TextLabel"))
        self.label_2.setText(_translate("figure8LoopAdd", "centre easting 1 (m)"))
        self.label.setText(_translate("figure8LoopAdd", "centre northing 1 (m)"))
        self.label_7.setText(_translate("figure8LoopAdd", "centre northing 2 (m)"))
        self.label_5.setText(_translate("figure8LoopAdd", "segments"))
        self.label_3.setText(_translate("figure8LoopAdd", "height (m)"))
        self.label_4.setText(_translate("figure8LoopAdd", "radius  (m)"))
        self.loopRadius.setToolTip(_translate("figure8LoopAdd", "<html><head/><body><p>Radius of the loop</p></body></html>"))



#################################################################################
# GJI final pub specs                                                           #
import matplotlib                                                               #
from matplotlib import rc                                                           #

#matplotlib.rcParams['text.latex.preamble']=[r"\usepackage{timet,amsmath,amssymb}"]  #
#rc('font',**{'family':'sans-serif','serif':['timet']})                                   #
#rc('font',**{'size':11})                                                             #
#rc('text', usetex=True)                                                             #

# converts pc that GJI is defined in to inches                                      # 
# In GEOPHYSICS \textwidth = 42pc                                               #
#        \columnwidth = 20pc                                                    #
#        one column widthe figures are 20 picas                                 #
#        one and one third column figures are 26 picas                          #
def pc2in(pc):                                                                  #
    return pc*12/72.27                                                          #
#################################################################################
import numpy as np
light_grey = np.array([float(248)/float(255)]*3)

def fixLeg(legend):
    rect = legend.get_frame()
    #rect.set_color('None')
    rect.set_facecolor(light_grey)
    rect.set_linewidth(0.0)
    rect.set_alpha(0.5)

def deSpine(ax1):
    spines_to_remove = ['top', 'right']
    for spine in spines_to_remove:
        ax1.spines[spine].set_visible(False)
    #ax1.xaxis.set_ticks_position('none')
    #ax1.yaxis.set_ticks_position('none')
    ax1.get_xaxis().tick_bottom()
    ax1.get_yaxis().tick_left()



from akvo.tressel.SlidesPlot import *
import numpy as np
import sys
import matplotlib.pyplot as plt
import cmocean
from pylab import meshgrid 
from akvo.tressel.logbarrier import *
import yaml,os

from matplotlib.colors import LogNorm
from matplotlib.colors import LightSource
from matplotlib.ticker import ScalarFormatter
from matplotlib.ticker import MaxNLocator
from matplotlib.ticker import AutoMinorLocator
from matplotlib.ticker import LogLocator
from matplotlib.ticker import FormatStrFormatter
from matplotlib.colors import Normalize

import cmocean
from akvo.tressel.lemma_yaml import * 

import pandas as pd

def buildKQT(K0,tg,T2Bins):
    """ 
        Constructs a QT inversion kernel from an initial amplitude one.  
    """
    nlay, nq = np.shape(K0)
    nt2 = len(T2Bins)
    nt = len(tg)
    KQT = np.zeros( ( nq*nt,nt2*nlay) )
    for iq in range(nq):
        for it in range(nt):
            for ilay in range(nlay):
                for it2 in range(nt2):
                    #KQT[iq*nt + it,ilay*nt2+it2] = K0[ilay,iq]*np.exp(-((10+tg[it])*1e-3)/(1e-3*T2Bins[it2]))
                    KQT[iq*nt + it,ilay*nt2+it2] = K0[ilay,iq]*np.exp(-((10+tg[it])*1e-3)/(1e-3*T2Bins[it2]))
    return KQT

def loadAkvoData(fnamein, chan):
    """ Loads data from an Akvo YAML file. The 0.02 is hard coded as the pulse length. This needs to be 
        corrected in future kernel calculations. The current was reported but not the pulse length. 
    """
    fname = (os.path.splitext(fnamein)[0])
    with open(fnamein, 'r') as stream:
        try:
            AKVO = (yaml.load(stream, Loader=yaml.Loader))
        except yaml.YAMLError as exc:
            print(exc)
            exit()

    Z = np.zeros( (AKVO.nPulseMoments, AKVO.Gated["Pulse 1"]["abscissa"].size ) )
    ZS = np.zeros( (AKVO.nPulseMoments, AKVO.Gated["Pulse 1"]["abscissa"].size ) )
    for q in range(AKVO.nPulseMoments):
        Z[q] = AKVO.Gated["Pulse 1"][chan]["Q-"+str(q) + " CA"].data
        if chan == "Chan. 1":
            ZS[q] = AKVO.Gated["Pulse 1"][chan]["STD"].data
        elif chan == "Chan. 2":
            ZS[q] = AKVO.Gated["Pulse 1"][chan]["STD"].data
        elif chan == "Chan. 3":
            ZS[q] = AKVO.Gated["Pulse 1"][chan]["STD"].data
        elif chan == "Chan. 4":
            ZS[q] = AKVO.Gated["Pulse 1"][chan]["STD"].data
        else:
            print("DOOM!!!")
            exit()
    #Z *= 1e-9 
    #ZS *= 1e-9 

    J = AKVO.Pulses["Pulse 1"]["current"].data 
    J = np.append(J,J[-1]+(J[-1]-J[-2]))
    Q = AKVO.pulseLength[0]*J
    return Z, ZS, AKVO.Gated["Pulse 1"]["abscissa"].data  #, Q

def catLayers(K0):
    K = np.zeros( (len(K0.keys()), len(K0["layer-0"].data)) , dtype=complex )
    for lay in range(len(K0.keys())):
        #print(K0["layer-"+str(lay)].data) #    print (lay)
        K[lay] =K0["layer-"+str(lay)].data #    print (lay)
    return 1e9*K                           # invert in nV

def loadK0(fname):
    """ Loads in initial amplitude kernel
    """
    print("loading K0", fname)
    with open(fname) as f:
        K0 = yaml.load(f, Loader=yaml.Loader)
    K = catLayers(K0.K0)
    ifaces = np.array(K0.Interfaces.data)
    return ifaces, np.abs(K)



def main():

    if (len (sys.argv) < 2):
        print ("akvoQT   invertParameters.yaml")
        exit()
    
    with open(sys.argv[1], 'r') as stream:
        try:
            cont = (yaml.load(stream, Loader=yaml.Loader))
        except yaml.YAMLError as exc:
            print(exc)

    ###############################################
    # Load in data
    ###############################################
    V = []
    VS = []
    tg = 0
    for dat in cont['data']:
        for ch in cont['data'][dat]['channels']:
            print("dat", dat, "ch", ch)
            v,vs,tg = loadAkvoData(dat, ch)
            V.append(v)
            VS.append(vs)
    for iv in range(1, len(V)):
        V[0] = np.concatenate( (V[0], V[iv]) )
        VS[0] = np.concatenate( (VS[0], VS[iv]) )
    V = V[0]
    VS = VS[0]

    ###############################################
    # Load in kernels
    ###############################################
    K0 = []
    for kern in cont["K0"]:
        ifaces,k0 = loadK0( kern )
        K0.append(k0)
    for ik in range(1, len(K0)):
        K0[0] = np.concatenate( (K0[0].T, K0[ik].T) ).T
    K0 = K0[0]
    #plt.matshow(K0)
    
    # VERY Simple DOI
    SNR = np.sum(.01*K0, axis=1) / VS[0][0]
    SNR[SNR>1] = 1
    SNRidx = 0 
    while SNR[SNRidx] >= 1:
        SNRidx += 1
    #print(SNR)
    #plt.plot(ifaces[0:-1], SNR)
    #plt.gca().axhline(y=VS[0][0], xmin=0, xmax=ifaces[-1], color='r')
    #plt.gca().axhline(y=1, xmin=0, xmax=ifaces[-1], color='r')
    #plt.show()

    ###############################################
    # Build full kernel
    ############################################### 
    T2Bins = np.logspace( np.log10(cont["T2Bins"]["low"]), np.log10(cont["T2Bins"]["high"]), cont["T2Bins"]["number"], endpoint=True, base=10)  
    KQT = buildKQT(K0,tg,T2Bins)
 
    ###############################################
    # Invert
    ############################################### 
    print("Calling inversion", flush=True)
    inv, ibreak, errn, phim, phid, mkappa = logBarrier(KQT, np.ravel(V), T2Bins, "lcurve", MAXITER=150, sigma=np.ravel(VS), alpha=1e6, smooth="Smallest" ) 


    ###############################################
    # Appraise
    ###############################################
 
    pre = np.dot(KQT,inv) 
    PRE = np.reshape( pre, np.shape(V)  )
    plt.matshow(PRE, cmap='Blues')
    plt.gca().set_title("predicted")
    plt.colorbar()

    DIFF = (PRE-V) / VS
    md = np.max(np.abs(DIFF))
    plt.matshow(DIFF, cmap=cmocean.cm.balance, vmin=-md, vmax=md)
    plt.gca().set_title("misfit / $\widehat{\sigma}$")
    plt.colorbar()
    
    plt.matshow(V, cmap='Blues')
    plt.gca().set_title("observed")
    plt.colorbar()

    T2Bins = np.append( T2Bins, T2Bins[-1] + (T2Bins[-1]-T2Bins[-2]) )
    
    INV = np.reshape(inv, (len(ifaces)-1,cont["T2Bins"]["number"]) )

    #alphas = np.tile(SNR, (len(T2Bins)-1,1))
    #colors = Normalize(1e-6, np.max(INV.T), clip=True)(INV.T)
    #colors = cmocean.cm.tempo(colors)
    ##colors[..., -1] = alphas
    #print(np.shape(colors)) 
    #print(np.shape(INV.T)) 

    #greys = np.full((*(INV.T).shape, 3), 70, dtype=np.uint8)

    Y,X = meshgrid( ifaces, T2Bins )
    fig = plt.figure( figsize=(pc2in(20.0),pc2in(22.)) )
    ax1 = fig.add_axes( [.2,.15,.6,.7] )
    im = ax1.pcolor(X, Y, INV.T, cmap=cmocean.cm.tempo) #cmap='viridis')
    #im = ax1.pcolor(X[0:SNRidx,:], Y[0:SNRidx,:], INV.T[0:SNRidx,:], cmap=cmocean.cm.tempo) #cmap='viridis')
    #im = ax1.pcolor(X[SNRidx::,:], Y[SNRidx::,:], INV.T[SNRidx::,:], cmap=cmocean.cm.tempo, alpha=.5) #cmap='viridis')
    #im = ax1.pcolormesh(X, Y, INV.T, alpha=alphas) #, cmap=cmocean.cm.tempo) #cmap='viridis')
    #im = ax1.pcolormesh(X, Y, INV.T, alpha=alphas) #, cmap=cmocean.cm.tempo) #cmap='viridis')
    #ax1.axhline( y=ifaces[SNRidx], xmin=T2Bins[0], xmax=T2Bins[-1], color='black'  )
    im.set_edgecolor('face')
    ax1.set_xlim( T2Bins[0], T2Bins[-1] )
    ax1.set_ylim( ifaces[-1], ifaces[0] )
    cb = plt.colorbar(im, label = u"PWC (m$^3$/m$^3$)") #, format='%1.1f')
    cb.locator = MaxNLocator( nbins = 4)
    cb.ax.yaxis.set_offset_position('left')                         
    cb.update_ticks()
 
    ax1.set_xlabel(u"$T_2^*$ (ms)")
    ax1.set_ylabel(u"depth (m)")
    
    ax1.get_xaxis().set_major_formatter(FormatStrFormatter('%1.0f'))
    ax1.get_yaxis().set_major_formatter(FormatStrFormatter('%1.0f'))
    ax1.xaxis.set_major_locator( MaxNLocator(nbins = 4) )   

    #ax1.xaxis.set_label_position('top') 

    ax2 = ax1.twiny()
    ax2.plot( np.sum(INV, axis=1), (ifaces[1:]+ifaces[0:-1])/2 ,  color='red' )
    ax2.set_xlabel(u"total water (m$^3$/m$^3$)")
    ax2.set_ylim( ifaces[-1], ifaces[0] )
    ax2.xaxis.set_major_locator( MaxNLocator(nbins = 3) )   
    ax2.get_xaxis().set_major_formatter(FormatStrFormatter('%0.2f'))
    ax2.axhline( y=ifaces[SNRidx], xmin=0, xmax=1, color='black', linestyle='dashed'  )
    #ax2.xaxis.set_label_position('bottom') 

    plt.savefig("akvoInversion.pdf")

    #############
    # water plot#

    fig2 = plt.figure( figsize=(pc2in(20.0),pc2in(22.)) )
    ax = fig2.add_axes( [.2,.15,.6,.7] )
    
    # Bound water cutoff 
    Bidx = T2Bins[0:-1]<33.0
    twater = np.sum(INV, axis=1)
    bwater = np.sum(INV[:,Bidx], axis=1)
    
    ax.plot( twater, (ifaces[0:-1]+ifaces[1::])/2, label="NMR total water", color='blue' )
    ax.plot( bwater, (ifaces[0:-1]+ifaces[1::])/2, label="NMR bound water", color='green' )
    
    ax.fill_betweenx((ifaces[0:-1]+ifaces[1::])/2 , twater, bwater, where=twater >= bwater, facecolor='blue', alpha=.5)
    ax.fill_betweenx((ifaces[0:-1]+ifaces[1::])/2 , bwater, 0, where=bwater >= 0, facecolor='green', alpha=.5)
    
    ax.set_xlabel(r"$\theta_N$ (m$^3$/m$^3$)")
    ax.set_ylabel(r"depth (m)")
    
    ax.set_ylim( ifaces[-1], ifaces[0] )
    ax.set_xlim( 0, ax.get_xlim()[1] )
    
    ax.axhline( y=ifaces[SNRidx], xmin=0, xmax=1, color='black', linestyle='dashed'  )
    
    plt.savefig("akvoInversionWC.pdf")
    plt.legend()  
 

    # Report results into a text file 
    fr = pd.DataFrame( INV, columns=T2Bins[0:-1] )
    fr.insert(0, "layer top", ifaces[0:-1] )
    fr.insert(1, "layer bottom", ifaces[1::] )
    fr.insert(2, "NMR total water", np.sum(INV, axis=1) )
    fr.insert(3, "NMR bound water", bwater )
    fr.insert(4, "Layer SNR", SNR )

    fr.to_csv("akvoInversion.csv")    
    #fr.to_excel("akvoInversion.xlsx")    

 
    plt.show()

if __name__ == "__main__":
    main() 




import yaml
    
def Complex_ctor(loader, node):
    re = eval(node.value[0][1].value)
    im = eval(node.value[1][1].value)
    return re + 1j*im 

yaml.add_constructor(r'Complex', Complex_ctor)

class MatrixXr(yaml.YAMLObject):
    yaml_tag = u'MatrixXr'
    def __init__(self, rows, cols, data):
        self.rows = rows
        self.cols = cols
        self.data = np.zeros((rows,cols))
    def __repr__(self):
        return "%s(rows=%r, cols=%r, data=%r)" % (self.__class__.__name__, self.rows, self.cols, self.data) 

class VectorXr(yaml.YAMLObject):
    yaml_tag = r'VectorXr'
    def __init__(self, array):
        self.size = np.shape(array)[0]
        self.data = array.tolist()
    def __repr__(self):
        # Converts to numpy array on import 
        return "np.array(%r)" % (self.data)

class VectorXcr(yaml.YAMLObject):
    yaml_tag = r'VectorXcr'
    def __init__(self, array):
        self.size = np.shape(array)[0]
        self.datar = array.tolist()
    def __repr__(self):
        # Converts to numpy array on import 
        #return "np.array(%r)" % (self.data)
        return "np.array(%r)" % (3)

class Vector3r(yaml.YAMLObject):
    yaml_tag = r'Vector3r'
    def __init__(self, array):
        self.size = 3 #np.shape(array)[0]
        self.data = array.tolist()
    def __repr__(self):
        # Converts to numpy array on import 
        return "np.array(%r)" % (self.data)

class Vector3Xcr(yaml.YAMLObject):
    yaml_tag = r'Vector3Xcr'
    def __init__(self, array):
        self.size = 3 #np.shape(array)[0]
        self.data = array.tolist()
    def __repr__(self):
        # Converts to numpy array on import 
        return "np.array(%r)" % (self.data)

class Vector3Xr(yaml.YAMLObject):
    yaml_tag = r'Vector3Xr'
    def __init__(self, array):
        self.size = 3 #np.shape(array)[0]
        self.data = array.tolist()
    def __repr__(self):
        # Converts to numpy array on import 
        return "np.array(%r)" % (self.data)

#class KernelV0( ):
    #yaml_tag = r'KernelV0'
#    def __init__(self):
#        self.name = "hello"

#def KernelV0_constructor(loader, node):
    #...     value = loader.construct_scalar(node)
    #...     a, b = map(int, value.split('d'))
#    return KernelV0( )


# class KervnelV0(yaml.YAMLObject):
#     yaml_loader = yaml.Loader
#     yaml_dumper = yaml.Dumper
# 
#     yaml_tag = u'!KernelV0'
#     #yaml_flow_style = ...
# 
#     def __init__(self):
#         self.val = 7
# 
#     @classmethod
#     def from_yaml(cls, loader, node):
#         # ...
#         data = 0
#         return data
# 
#     @classmethod
#     def to_yaml(cls, dumper, data):
#         # ...
#         return node

class KervnelV0(yaml.YAMLObject):
    yaml_tag = u'KernelV0'
    def __init__(self, val):
        self.val = val

class LayeredEarthEM(yaml.YAMLObject):
    yaml_tag = u'LayeredEarthEM'
    def __init__(self, val):
        self.val = val

class PolygonalWireAntenna(yaml.YAMLObject):
    yaml_tag = u'PolygonalWireAntenna'
    def __init__(self, val):
        self.val = val

class AkvoData(yaml.YAMLObject):
    yaml_tag = u'AkvoData'
    def __init__(self, obj): #nPulseMoments, pulseLength):
    #def __init__(self, rows, cols, data):
        #self.nPulseMoments = nPulseMoments
        #self.pulseLength = pulseLength 
        #for key in obj.keys:
        #    self[key] = obj.key
        pass

